package org.briarproject.bramble.plugin.lora;

import org.briarproject.bramble.plugin.lora.DeviceManager;
import org.briarproject.bramble.plugin.lora.L2PWriter;
import org.briarproject.bramble.plugin.lora.L2PReader;
import org.briarproject.bramble.plugin.lora.L2PReaderFactory;
import org.briarproject.bramble.api.Pair;
import org.briarproject.nullsafety.NotNullByDefault;
import org.briarproject.bramble.api.plugin.ConnectionHandler;
import org.briarproject.bramble.api.plugin.PluginCallback;
import org.briarproject.bramble.api.plugin.Backoff;
import org.briarproject.bramble.api.plugin.TransportConnectionReader;
import org.briarproject.bramble.api.plugin.TransportConnectionWriter;
import org.briarproject.bramble.api.plugin.TransportId;
import org.briarproject.bramble.api.plugin.simplex.SimplexPlugin;
import org.briarproject.bramble.api.properties.TransportProperties;
import org.briarproject.bramble.api.system.Clock;
import org.briarproject.bramble.api.plugin.PluginException;
import org.briarproject.bramble.api.lifecycle.IoExecutor;
import java.util.concurrent.Executor;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.Collection;
import java.util.logging.Logger;

import javax.annotation.concurrent.Immutable;

import static java.util.Collections.singletonMap;
import static java.util.logging.Level.WARNING;
import static java.util.logging.Logger.getLogger;
import static org.briarproject.bramble.api.plugin.Plugin.State.ACTIVE;
import static org.briarproject.bramble.api.plugin.LoRaConstants.ID;
import static org.briarproject.bramble.api.plugin.file.RemovableDriveConstants.PROP_SUPPORTED; //TODO
import static org.briarproject.bramble.util.LogUtils.logException;

@Immutable
@NotNullByDefault
abstract class AbstractLoRaPlugin implements SimplexPlugin/*, EventHandler, EventListener*/ {

	private static final Logger LOG =
			getLogger(AbstractLoRaPlugin.class.getName());

	protected final Clock clock;
	protected final Backoff backoff; //TODO: Is this needed?
	protected final long maxLatency;
	protected final PluginCallback callback;
	protected final int maxIdleTime;
	protected DeviceManager devicemanager;
	protected final Executor ioExecutor, wakefulIoExecutor;
	protected L2PReaderFactory rf;

	AbstractLoRaPlugin(
		Executor ioExecutor,
		Executor wakefulIoExecutor, Clock clock, Backoff backoff, PluginCallback callback, long maxLatency, int maxIdleTime) {
		LOG.info("AbstractLoRaPlugin::Constructor");
		this.clock = clock;
		this.callback = callback;
		this.backoff = backoff;
		this.ioExecutor = ioExecutor;
		this.wakefulIoExecutor = wakefulIoExecutor;
		this.maxLatency = maxLatency;
		this.maxIdleTime = maxIdleTime;

		this.rf = new L2PReaderFactory(callback);
		this.devicemanager = new DeviceManager();
	}

	public LoRaAdapter getAdapter() {
		return devicemanager.getAdapter();
	}

	public boolean TXenabled() {
		return true;
	}

	@Override
	public TransportId getId() {
		return ID;
	}

	@Override
	public long getMaxLatency() {
		return maxLatency;
	}

	@Override
	public int getMaxIdleTime() {
		// Unused for simplex transports
		throw new UnsupportedOperationException();
	}

	@Override
	public boolean isLossyAndCheap() {
		return false;
	}

	@Override
	public State getState() {
		LOG.info("getState(): "+devicemanager.getState());
		return devicemanager.getState();
	}

	@Override
	public void start() throws PluginException {
		LOG.info("start()");
		devicemanager.getAdapter().start();
		callback.pluginStateChanged(getState());
	}

	@Override
	public void stop() throws PluginException {
		LOG.info("stop()");
		devicemanager.getAdapter().stop();
	}

	//TODO
	@Override
	public int getReasonsDisabled() {
		return 0;
	}

	@Override
	public boolean shouldPoll() {
		return true;
	}

	@Override
	public int getPollingInterval() {
		LOG.info("getPollingInterval(): "+backoff.getPollingInterval());
		return backoff.getPollingInterval();
	}

	//TODO: Switch to Events
	@Override
	public void poll(Collection<Pair<TransportProperties, ConnectionHandler>> properties) {
		//backoff.increment();
		LOG.info("poll()");
		if (getState() != ACTIVE || !TXenabled()) return;
		backoff.increment();
		for (Pair<TransportProperties, ConnectionHandler> p : properties) {
			wakefulIoExecutor.execute(() -> {
				//TODO: Is there anything to send?
				TransportConnectionWriter w = new L2PWriter(this);
				if (w != null) {
					backoff.reset();
					p.getSecond().handleWriter(w);
				}
			});
		}
	}

	@Override
	public TransportConnectionReader createReader(TransportProperties p) {
		//TODO: Where is that called?
		LOG.info("createReader()");
		//callback.handleConnection(conn); used
		return null;
		/*try {
			return new TransportInputStreamReader(openInputStream(p));
		} catch (IOException e) {
			logException(LOG, WARNING, e);
			return null;
		}*/
	}

	@Override
	public TransportConnectionWriter createWriter(TransportProperties p) {
		//TODO: Where is that called?
		LOG.info("createWriter()");
		return new L2PWriter(this);
	}
}
