package org.briarproject.bramble.plugin.lora;

import org.briarproject.bramble.api.battery.BatteryManager;
import org.briarproject.bramble.api.crypto.CryptoComponent;
import org.briarproject.bramble.api.event.EventBus;
import org.briarproject.bramble.api.lifecycle.IoExecutor;
import org.briarproject.bramble.api.network.NetworkManager;
import org.briarproject.nullsafety.NotNullByDefault;
import org.briarproject.bramble.api.plugin.Backoff;
import org.briarproject.bramble.api.plugin.BackoffFactory;
import org.briarproject.bramble.api.plugin.PluginCallback;
import org.briarproject.bramble.api.plugin.TransportId;
import org.briarproject.bramble.api.plugin.simplex.SimplexPlugin;
import org.briarproject.bramble.api.plugin.simplex.SimplexPluginFactory;
import org.briarproject.bramble.api.plugin.LoRaConstants;
import org.briarproject.bramble.api.system.Clock;
import org.briarproject.bramble.api.system.LocationUtils;
import org.briarproject.bramble.api.system.ResourceProvider;
import org.briarproject.bramble.api.system.WakefulIoExecutor;

import java.io.File;
import java.util.concurrent.Executor;
import java.util.logging.Logger;

import javax.annotation.Nullable;
import javax.annotation.concurrent.Immutable;
import javax.net.SocketFactory;

import static java.util.logging.Level.INFO;
import static java.util.logging.Logger.getLogger;

@Immutable
@NotNullByDefault
abstract class AbstractLoRaPluginFactory implements SimplexPluginFactory {

	protected static final Logger LOG =
			getLogger(AbstractLoRaPluginFactory.class.getName());

	protected final BackoffFactory backoffFactory;
	protected final Clock clock;
	protected final CryptoComponent crypto;
	protected final Executor ioExecutor, wakefulIoExecutor;

	AbstractLoRaPluginFactory(Executor ioExecutor,
	Executor wakefulIoExecutor,BackoffFactory backoffFactory, Clock clock, CryptoComponent crypto) {
		this.backoffFactory = backoffFactory;
		this.clock = clock;
		this.crypto = crypto;
		this.ioExecutor = ioExecutor;
		this.wakefulIoExecutor = wakefulIoExecutor;
	}

	abstract AbstractLoRaPlugin createPluginInstance(Backoff backoff, PluginCallback callback);

	@Override
	public TransportId getId() {
		return LoRaConstants.ID;
	}

	@Override
	public long getMaxLatency() {
		return LoRaConstants.MAX_LATENCY;
	}

	@Override
	public SimplexPlugin createPlugin(PluginCallback callback) {

		Backoff backoff = backoffFactory.createBackoff(LoRaConstants.MIN_POLLING_INTERVAL, LoRaConstants.MAX_POLLING_INTERVAL, LoRaConstants.BACKOFF_BASE);
		AbstractLoRaPlugin plugin = createPluginInstance(backoff, callback);
		//eventBus.addListener(plugin);
		return plugin;
	}
}
