package org.briarproject.bramble.plugin.lora;

import org.briarproject.bramble.plugin.lora.AbstractLoRaPlugin;
import org.briarproject.bramble.api.plugin.TransportConnectionReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.PipedOutputStream;
import java.io.PipedInputStream;
import org.briarproject.bramble.plugin.lora.LoRaAdapter;
import org.briarproject.bramble.plugin.lora.L2PReaderFactory;
import org.briarproject.bramble.util.ByteUtils;
import java.util.Arrays;
import java.lang.System;
import org.briarproject.bramble.api.Bytes;
import java.util.logging.Logger;
import java.util.concurrent.LinkedBlockingDeque;
import java.lang.InterruptedException;
import static org.briarproject.bramble.api.transport.TransportConstants.TAG_LENGTH;
import static org.briarproject.bramble.api.plugin.LoRaConstants.*;

public class L2PReader implements TransportConnectionReader {
    private L2PReaderFactory f;
    private L2PInputStream stream = new L2PInputStream(this);
    private Long nextSequenceNr = 0L;

	private static final Logger LOG =
			Logger.getLogger(L2PReader.class.getName());

    public L2PReader(L2PReaderFactory factory) {
        f = factory;
    }

    public void parse(byte content[], Bytes tag, long sequenceNR) {
        LOG.info("reader pkg "+sequenceNR+" with tag "+ByteUtils.bytesToHex(tag.getBytes())+": "+ByteUtils.bytesToHex(content));

        try {
            //lost a Packet
            if(sequenceNR != nextSequenceNr) {
                //we don't have a better mechanism :(
                LOG.info("read sequenceNr " + sequenceNR + " expected "+nextSequenceNr);
                this.dispose(false, true);
            }

            if(sequenceNR == 0) {
                System.arraycopy(tag.getBytes(),0,content,0,TAG_LENGTH);
                stream.add(content);
            } else {
                stream.add(Arrays.copyOfRange(content,TAG_LENGTH,PKG_SIZE));
            }

            nextSequenceNr++;
        } catch (IOException e) {
            // TODO
        }
    }

    @Override
    public InputStream getInputStream() throws IOException {
        return stream;
    }

    @Override
    public void dispose(boolean exception, boolean recognised)
            throws IOException {
        f.remove(this);
    }

    //TODO: Syncronize
    private class L2PInputStream extends InputStream {
        private L2PReader r;
        private LinkedBlockingDeque<Bytes> buffer = new LinkedBlockingDeque<>();
        private int offset = 0;
        private boolean closed=false;

        public L2PInputStream(L2PReader r) {
            this.r = r;
        }

        public void add(byte n[]) {
            buffer.offerLast(new Bytes(n));
        }

        //just use the parent wrapper
        //@Override
        //public int read(byte[] b) throws IOException

        //also just use parent wrapper
        //public long skip(long n) throws IOException

        @Override
        public int read() throws IOException {
            if(buffer.size() == 0 && closed)
                return -1;
            int ret=0;
            try {
                Bytes block = buffer.take();
                ret = block.getBytes()[offset++];
                if(offset == block.getBytes().length)
                    offset = 0;
                    //next Block
                else
                    buffer.putFirst(block);
            } catch(InterruptedException e) {
            }
            
            return ret;
        }

        @Override
        public int read(byte[] b, int off, int len) throws IOException {
            if(buffer.size() == 0 && closed) {
                return -1;
            }

            int read=0;
            try {
                while(len > 0 && (buffer.size() > 0 || read == 0)) {
                    Bytes block = buffer.take();
                    int l = Math.min(len, block.getBytes().length-offset);
                    System.arraycopy(block.getBytes(), offset, b, off, l);
                    offset += l;
                    off += l;
                    read += l;
                    len -= l;
                    if(offset >= block.getBytes().length)
                        offset = 0;
                        //discard block from buffer
                    else
                        buffer.putFirst(block);
                }
            } catch(InterruptedException e) {
            }

            return read;
        }

        @Override
        public int available() throws IOException {
            return (buffer.size()>0 && !closed)?PKG_SIZE-TAG_LENGTH:0;
        }

        @Override
        public void close() throws IOException {
            closed=true;
            buffer.offerLast(new Bytes(new byte[1])); //push dummy data (like padding) to the Reader to not lock the reader
            r.dispose(false,false);
        }

        //TODO: needed?
        @Override
        public boolean markSupported() {
            return false;
        }
    }
}