package org.briarproject.bramble.plugin.lora;

import static org.briarproject.bramble.api.transport.TransportConstants.TAG_LENGTH;
import org.briarproject.bramble.plugin.lora.AbstractLoRaPlugin;
import org.briarproject.bramble.api.plugin.TransportConnectionReader;
import java.io.IOException;
import java.io.InputStream;
import org.briarproject.bramble.plugin.lora.LoRaAdapter;
import org.briarproject.bramble.plugin.lora.L2PReader;
import org.briarproject.bramble.plugin.lora.L2PTagManager;
import java.util.Arrays;
import java.lang.Integer;
import java.util.HashMap;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Collections;
import java.util.logging.Logger;
import org.briarproject.bramble.util.ByteUtils;
import org.briarproject.bramble.api.Bytes;
import org.briarproject.bramble.api.Pair;
import org.briarproject.bramble.api.plugin.PluginCallback;
import org.briarproject.bramble.api.plugin.LoRaConstants;

//TODO: Make singleton
public class L2PReaderFactory {
    //TODO: Do I need syncronisation here?
    private Map<Bytes, L2PReader> readers = Collections.synchronizedMap(new HashMap<Bytes, L2PReader>());

    private PluginCallback callback;
	private static final Logger LOG =
            Logger.getLogger(L2PReaderFactory.class.getName());
        
    private final L2PTagManager l2pTags;

    public L2PReaderFactory(PluginCallback callback) {
        this.callback = callback;
        l2pTags = L2PTagManager.TODOuseDependencyInjection(LoRaConstants.ID);
        l2pTags.setOpenReaders(readers);
    }

    private L2PReader createReader() {
        L2PReader r = new L2PReader(this);
        callback.handleReader(r);
        return r;
    }

    private L2PReader getReaderByBTPtag(Bytes btpTag) {
        L2PReader r = readers.get(btpTag);
        if(r == null) {
            r = createReader();
            readers.put(btpTag, r);
        }

        return r;
    }

    public void parse(byte[] content) {
        byte[] l2pTag = Arrays.copyOf(content, TAG_LENGTH);
        Pair<Bytes,Long> tagData = l2pTags.get(new Bytes(l2pTag));
        if(tagData == null) {
            //this packet isn't recognised => it probably is no briar packet => ignore
            LOG.info("dropping LoRa pkg "+ByteUtils.bytesToHex(content));
            LOG.info("Tags:"+l2pTags.toString());
            return;
        }
        getReaderByBTPtag(tagData.getFirst()).parse(content, tagData.getFirst(), tagData.getSecond().intValue());

        //TODO: Drop Tag if successfull!

        //generate next Tags
        l2pTags.addNfromBTPtag(tagData.getFirst(), tagData.getSecond()+1, L2PTagManager.CALCULATE_OPEN_CONN);
    }

    public void remove(L2PReader r) {
        for(Entry<Bytes,L2PReader> i: readers.entrySet())
            if(r.equals(i.getValue())) {
                Bytes BTPtag = i.getKey();
                readers.remove(BTPtag);
                l2pTags.removeAllFromBTPtag(BTPtag);
                return;
            }
    }
}
