package org.briarproject.bramble.plugin.lora;

import static org.briarproject.bramble.api.transport.TransportConstants.TAG_LENGTH;
import org.briarproject.bramble.plugin.lora.AbstractLoRaPlugin;
import org.briarproject.bramble.api.plugin.TransportConnectionReader;
import java.io.IOException;
import java.io.InputStream;
import org.briarproject.bramble.plugin.lora.LoRaAdapter;
import org.briarproject.bramble.plugin.lora.L2PReader;
import java.util.Arrays;
import java.lang.Long;
import java.util.HashMap;
import java.util.Map;
import java.lang.System;
import java.util.Collections;
import java.util.logging.Logger;
import org.briarproject.bramble.util.ByteUtils;
import org.briarproject.bramble.api.Bytes;
import org.briarproject.bramble.api.Pair;
import org.briarproject.bramble.api.plugin.PluginCallback;
import org.briarproject.bramble.api.plugin.TransportId;
import org.briarproject.bramble.api.plugin.LoRaConstants;
import org.briarproject.bramble.plugin.lora.L2PReader;
import org.briarproject.bramble.api.crypto.TransportCrypto;
import java.util.Map.Entry;
import java.util.Iterator;
import java.util.Collection;

import org.bouncycastle.crypto.CryptoException;
import org.bouncycastle.crypto.Digest;
import org.bouncycastle.crypto.digests.Blake2bDigest;

import static java.util.concurrent.TimeUnit.MILLISECONDS;
import static java.util.logging.Level.WARNING;
import static java.util.logging.Logger.getLogger;
import static org.briarproject.nullsafety.NullSafety.requireExactlyOneNull;
import static org.briarproject.bramble.api.transport.TransportConstants.MAX_CLOCK_DIFFERENCE;
import static org.briarproject.bramble.api.transport.TransportConstants.PROTOCOL_VERSION;
import static org.briarproject.bramble.api.transport.TransportConstants.TAG_LENGTH;
import static org.briarproject.bramble.util.LogUtils.logException;
import static org.briarproject.bramble.util.ByteUtils.INT_16_BYTES;
import static org.briarproject.bramble.util.ByteUtils.INT_64_BYTES;
import static org.briarproject.bramble.util.ByteUtils.MAX_16_BIT_UNSIGNED;
import static org.briarproject.bramble.util.ByteUtils.MAX_32_BIT_UNSIGNED;
import static org.briarproject.bramble.util.ByteUtils.writeUint16;
import static org.briarproject.bramble.util.ByteUtils.writeUint64;
import java.util.logging.Logger;



public class L2PTagManager {

	private static final Logger LOG =
            Logger.getLogger(L2PTagManager.class.getName());
    

    //TODO: For better performance, we should choose a data-structure that is indexed by B2PTag as well as L2PTag
    private final Map<Bytes, Pair<Bytes,Long>> l2pTags = Collections.synchronizedMap(new HashMap<Bytes, Pair<Bytes,Long>>());
    private Map<Bytes, L2PReader> openReaders;
//    private TransportCrypto crypto; //move genL2PTag to TransportCrypto

    public String toString() {
        return l2pTags.toString();
    }

    //TODO: Evaluate what makes sense using a simple or sophisticated protocol
    public static final int CALCULATE_STD = 4;
    public static final int CALCULATE_OPEN_CONN = 16;

    public L2PTagManager() {
        //Init crypto with TransportCryptoImpl including dependency injection
    }

    public void addNfromBTPtag(final Bytes inTag, long start, int n) {
        for(long i=start; i<start+n; i++) {
            l2pTags.put(genL2PTag(inTag, i), new Pair<Bytes,Long>(inTag, i));
        }
    }

    //Bytes: BTP-Tag, Integer: sequenceNR
    public Pair<Bytes,Long> get(Bytes L2PTag) {
        return l2pTags.get(L2PTag);
    }

    public void remove(Bytes l2pTag) {
        l2pTags.remove(l2pTag);
    }

    public void removeAllFromBTPtag(Bytes inTag) {
        if(openReaders.get(inTag) != null)
            return;

        //TODO: More elegant option to remove Elements using lambda?
        Iterator<Pair<Bytes, Long>> i = l2pTags.values().iterator();
        while(i.hasNext())
            if(inTag.equals(i.next().getSecond()))
                i.remove();
    }

    public void clear() {
        l2pTags.clear();
    }

    public void setOpenReaders(Map<Bytes, L2PReader> rs) {
        this.openReaders = rs;
    }

    public static Bytes genL2PTag(final Bytes inTag, long i) {
        //I don't understand the 32 Bytes, I just copied it from TransportCryptoImpl ...
        Digest prf = new Blake2bDigest(inTag.getBytes(), 32, null, null);
		// The output of the PRF must be long enough to use as a tag
		int macLength = prf.getDigestSize();
		if (macLength < TAG_LENGTH) throw new IllegalStateException();

//		byte[] protocolVersionBytes = new byte[INT_16_BYTES];
//		writeUint16(protocolVersion, protocolVersionBytes, 0);
//		prf.update(protocolVersionBytes, 0, protocolVersionBytes.length);

        //TODO: could break a stream after 2^62 frames - would be Zettabytes of data
		byte[] sequenceNrBuffer = new byte[INT_64_BYTES];
		writeUint64(i, sequenceNrBuffer, 0);
		prf.update(sequenceNrBuffer, 0, sequenceNrBuffer.length);
		byte[] mac = new byte[macLength];
        prf.doFinal(mac, 0);
        
		// The output is the first TAG_LENGTH bytes of the MAC
        byte[] tag = new byte[TAG_LENGTH];
        System.arraycopy(mac, 0, tag, 0, TAG_LENGTH);
        return new Bytes(tag);
    }

    //TODO: Move to Dependency Injection
    private static final L2PTagManager m = new L2PTagManager();
    public static L2PTagManager TODOuseDependencyInjection(TransportId t) {
        return m;
    }
}