package org.briarproject.bramble.plugin.lora.adapter;

import org.briarproject.bramble.plugin.lora.LoRaAdapter;
import org.briarproject.bramble.plugin.lora.L2PReaderFactory;

import java.lang.System;
import java.security.SecureRandom;
import java.util.Arrays;

import com.sun.jna.Library;
import com.sun.jna.Native;
import com.sun.jna.Platform;
import com.sun.jna.Callback;
import com.sun.jna.Pointer;
import org.briarproject.bramble.util.ByteUtils;

import javax.inject.Inject;
import java.util.logging.Logger;

public class Pinedio implements LoRaAdapter {
	private final SecureRandom secureRandom;
	private L2PReaderFactory rf;

	private static final Logger LOG =
			Logger.getLogger(Pinedio.class.getName());


    private CLibrary.appRxCallback_ RxCallback = new CLibrary.appRxCallback_() {
		public void invoke(Pointer payload, byte RssiPkt, byte SnrPkt, byte SignalRssiPkt) {
            byte[] b = payload.getByteArray(0,255);
			LOG.info("Rx "+ByteUtils.bytesToHex(b));
			
			rf.parse(b);
        }
    };

    private interface CLibrary extends Library {
		CLibrary INSTANCE = (CLibrary)Native.loadLibrary("pinedio-lora-driver", CLibrary.class);

        interface appRxCallback_ extends Callback {
            void invoke(Pointer payload, byte RssiPkt, byte SnrPkt, byte SignalRssiPkt);
        }

        void PinedioLoraRadioInitialize(appRxCallback_ RxCallback);
        void PinedioLoraRadioSend(byte payload[]); //make sure we have exactly 255 elements!!
        void PinedioLoraRadioStart();
        void PinedioLoraRadioStop();
	}
	
//	@Inject
	public Pinedio(L2PReaderFactory rf) {
		this.rf = rf;

		this.secureRandom = new SecureRandom(); //TODO: Dependency Inection
		System.setProperty("jna.debug_load", "true");
		System.setProperty("jna.library.path", "/root/Bakk2/briar/");
		
		CLibrary.INSTANCE.PinedioLoraRadioInitialize(RxCallback);
		//TODO: Stop
	}

	public void start() {
		CLibrary.INSTANCE.PinedioLoraRadioStart();
	}

	public void stop() {
		CLibrary.INSTANCE.PinedioLoraRadioStop();
	}

	public void send(byte payload[]) {
		//padding - the end of Stream is already known at a higher level so we can just add random chars
		int length = payload.length;
		if(length < 255) {
			byte[] padding = new byte[255-length];
			secureRandom.nextBytes(padding);

			payload = Arrays.copyOf(payload, 255);
			System.arraycopy(padding, 0, payload, length, 255-length);
		}

		LOG.info("Tx " + ByteUtils.bytesToHex(payload));

		CLibrary.INSTANCE.PinedioLoraRadioSend(payload);
	}

	//TODO: Destructor
}